// Copyright 2015 The go-ethereum Authors
// This file is part of the go-ethereum library.
//
// The go-ethereum library is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// The go-ethereum library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with the go-ethereum library. If not, see <http://www.gnu.org/licenses/>.

// Package tests implements execution of Ethereum JSON tests.
package tests

import (
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"math/big"
	"os"
	"path/filepath"

	"github.com/eth-classic/go-ethereum/core"
)

var (
	baseDir            = filepath.Join(".", "files")
	blockTestDir       = filepath.Join(baseDir, "BlockchainTests")
	stateTestDir       = filepath.Join(baseDir, "StateTests")
	transactionTestDir = filepath.Join(baseDir, "TransactionTests")
	vmTestDir          = filepath.Join(baseDir, "VMTests")
	rlpTestDir         = filepath.Join(baseDir, "RLPTests")
	ethDir             = filepath.Join(".", "testData")
	ethGeneralStateDir = filepath.Join(ethDir, "GeneralStateTests")
	ethBasicTestDir    = filepath.Join(ethDir, "BasicTests")

	BlockSkipTests = initBlockSkipTests()

	/* Go client does not support transaction (account) nonces above 2^64. This
	technically breaks consensus but is regarded as "reasonable
	engineering constraint" as accounts cannot easily reach such high
	nonce values in practice
	*/
	TransSkipTests = []string{"TransactionWithHihghNonce256"}
	StateSkipTests = []string{}
	VmSkipTests    = []string{}
)

func initBlockSkipTests() []string {
	if core.UseSputnikVM == "true" {
		return []string{
			// These tests are not valid, as they are out of scope for RLP and
			// the consensus protocol.
			"BLOCK__RandomByteAtTheEnd",
			"TRANSCT__RandomByteAtTheEnd",
			"BLOCK__ZeroByteAtTheEnd",
			"TRANSCT__ZeroByteAtTheEnd",

			"ChainAtoChainB_blockorder2",
			"ChainAtoChainB_blockorder1",
			"ChainAtoChainB_BlockHash",
			"CallingCanonicalContractFromFork_CALLCODE",
		}
	} else {
		return []string{
			// These tests are not valid, as they are out of scope for RLP and
			// the consensus protocol.
			"BLOCK__RandomByteAtTheEnd",
			"TRANSCT__RandomByteAtTheEnd",
			"BLOCK__ZeroByteAtTheEnd",
			"TRANSCT__ZeroByteAtTheEnd",

			"ChainAtoChainB_blockorder2",
			"ChainAtoChainB_blockorder1",
		}
	}
}

func readJson(reader io.Reader, value interface{}) error {
	data, err := ioutil.ReadAll(reader)
	if err != nil {
		return fmt.Errorf("error reading JSON file: %v", err)
	}
	if err = json.Unmarshal(data, &value); err != nil {
		if syntaxerr, ok := err.(*json.SyntaxError); ok {
			line := findLine(data, syntaxerr.Offset)
			return fmt.Errorf("JSON syntax error at line %v: %v", line, err)
		}
		return fmt.Errorf("JSON unmarshal error: %v", err)
	}
	return nil
}

func readJsonFile(fn string, value interface{}) error {
	file, err := os.Open(fn)
	if err != nil {
		return err
	}
	defer file.Close()

	err = readJson(file, value)
	if err != nil {
		return fmt.Errorf("%s in file %s", err.Error(), fn)
	}
	return nil
}

// findLine returns the line number for the given offset into data.
func findLine(data []byte, offset int64) (line int) {
	line = 1
	for i, r := range string(data) {
		if int64(i) >= offset {
			return
		}
		if r == '\n' {
			line++
		}
	}
	return
}

// Forks table defines supported forks and their chain config.
var Forks = map[string]RuleSet{
	"Frontier": {},
	"Homestead": {
		HomesteadBlock: big.NewInt(0),
	},
	"EIP150": {
		HomesteadBlock:           big.NewInt(0),
		HomesteadGasRepriceBlock: big.NewInt(0),
	},
	"Byzantium": {
		HomesteadBlock:           big.NewInt(0),
		HomesteadGasRepriceBlock: big.NewInt(0),
		DiehardBlock:             big.NewInt(0),
		AtlantisBlock:            big.NewInt(0),
	},
}

// ChainConfigs table used to map configs to difficulty test files
var ChainConfigs = map[string]core.ChainConfig{
	// "difficulty":              {},
	"difficultyFrontier.json": {},
	"difficultyHomestead.json": {
		Forks: []*core.Fork{
			{
				Name:  "Homestead",
				Block: big.NewInt(0),
				Features: []*core.ForkFeature{
					{
						ID: "difficulty",
						Options: core.ChainFeatureConfigOptions{
							"type": "homestead",
						},
					},
					{
						ID: "gastable",
						Options: core.ChainFeatureConfigOptions{
							"type": "homestead",
						},
					},
				},
			},
		},
	},
	"difficultyByzantium.json": {
		Forks: []*core.Fork{
			{
				Name:  "Atlantis",
				Block: big.NewInt(0),
				Features: []*core.ForkFeature{
					{
						ID: "difficulty",
						Options: core.ChainFeatureConfigOptions{
							"type":   "atlantis",
							"length": 3000000,
						},
					},
				},
			},
		},
	},
}

// UnsupportedForkError is returned when a test requests a fork that isn't implemented.
type UnsupportedForkError struct {
	Name string
}

func (e UnsupportedForkError) Error() string {
	return fmt.Sprintf("unsupported fork %q", e.Name)
}
